/***************************************************************************
***
***    Copyright 2007  Hon Hai Precision Ind. Co. Ltd.
***    All Rights Reserved.
***    No portions of this material shall be reproduced in any form without the
***    written permission of Hon Hai Precision Ind. Co. Ltd.
***
***    All information contained in this document is Hon Hai Precision Ind.  
***    Co. Ltd. company private, proprietary, and trade secret property and 
***    are protected by international intellectual property laws and treaties.
***
****************************************************************************/

/*******************************************************************************
 ***  Filename:
 ***      wpsd.c
 ***
 ***  Description:
 ***      Control WPS LED status
 ***
 ***  History:
 ***
 ***   Modify Reason                    Author          Date         Search Flag(Option)
 ***--------------------------------------------------------------------------------------
 ***   Initial version                  pling           07/16/007 
 ******************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <sys/ioctl.h>

#include <epivers.h>
#include <shutils.h>
#include <bcmdevs.h>
#include <bcmparams.h>
#include <bcmgpio.h>


#include "wps_led.h"

#define GPIO_WPS_BUTTON         6
#define GPIO_11N_ANT0_TX        3
#define GPIO_11N_ANT1_TX        2
#define GPIO_11G_ANT            4

#define DEV_WPS_LED             "/dev/wps_led"

#define ACTION_NONE             0
#define ACTION_BLINK_TOGGLE     1
#define ACTION_BLINK_WPS        2

#define LOCAL       static

static int check_wps_button(void)
{
    unsigned long value;
    int i;
    
    bcmgpio_in(1 << GPIO_WPS_BUTTON, &value);
    if (value)
        return ACTION_NONE;

    /* Wait up to 3 second for user load default */
    for (i=0; i<30; i++)
    {
        bcmgpio_in(1 << GPIO_WPS_BUTTON, &value);

        /*
         * If button is released before the 3sec mark, 
         * toggle LED on / off
         */
        if (value)
            return ACTION_BLINK_TOGGLE;

        usleep(100000);
    }
    
    /* 3 sec passed, start WPS !!! */
    return ACTION_BLINK_WPS;
}

static int check_ant_status(void)
{
    unsigned long value;
   
    /* Read all 3 ant Inputs at the same time */
    bcmgpio_in(7L << GPIO_11N_ANT1_TX, &value);

    value >>= GPIO_11N_ANT1_TX;

    return (int)value;
}

LOCAL int convert_ant_status_to_LED_pattern(int ant_status)
{
    /*
     * The following conversion of ANT status to LED patten follows
     * NETGEAR Patrick Rata's spec:
     *
     * 11n ant0 + 11n ant1  : blink LED 2, 4, 6, 8
     * 11n ant0 only        : blink LED 1, 4, 5, 8
     * 11n ant1 only        : blink LED 1, 2, 5, 6
     * 
     * 11g ant on           : blink LED 7
     * 11g ant off          : blink LED 3
     */

    /*
     * There is a output register to control the 8 LEDs:
     * bit 0 : LED 1, 5
     * bit 1 : LED 2, 6
     * bit 2 : LED 4, 8
     * bit 3 : LED 3
     * bit 4 : LED 7
     */
    
    int blink_pattern = 0;
    int ant0_11n, ant1_11n, ant_11g;

    ant0_11n = ant_status & 0x2;
    ant1_11n = ant_status & 0x1;
    ant_11g  = ant_status & 0x4;

    if (ant0_11n && ant1_11n)
        blink_pattern = 0x02 & 0x04;        /* bit 1 + bit 2 */
    else
    if (ant0_11n)
        blink_pattern = 0x01 & 0x04;        /* bit 0 + bit 2 */
    else
    if (ant1_11n)
        blink_pattern = 0x01 & 0x02;        /* bit 0 + bit 1 */

    if (ant_11g)
        blink_pattern |= 0x10;              /* bit 4 */
    else
        blink_pattern |= 0x08;              /* bit 3 */
    
    return blink_pattern;
}

int send_wps_led_cmd(int cmd, int arg)
{
    int ret_val=0;
    int fd;

    fd = open(DEV_WPS_LED, O_RDWR);
    if (fd < 0) 
        return -1;

    ret_val = ioctl(fd, cmd, arg);
    close(fd);

    return ret_val;
}

int main(int argc, char *argv[])
{
    int status;
    int loop_count = 0;

    int wps_start_count=0;
    int wps_led_state = 1;
    int ant_status;
    int wps_led_pattern;
    
    /* Daemonize myself */
    if (daemon (1, 1) == -1)
        perror ("daemon");

    /* Setup GPIOs for reading */
    bcmgpio_connect(GPIO_WPS_BUTTON,  BCMGPIO_DIRN_IN);
    bcmgpio_connect(GPIO_11N_ANT0_TX, BCMGPIO_DIRN_IN);
    bcmgpio_connect(GPIO_11N_ANT1_TX, BCMGPIO_DIRN_IN);
    bcmgpio_connect(GPIO_11G_ANT,     BCMGPIO_DIRN_IN);

    while (1)
    {
        /* Check if WPS is start */
        if (wps_start_count != 0)
        {
            /* Check whether WPS should time out now */
            if (loop_count - wps_start_count >= 480)
            {
                /* WPS time out in 2 minutes */
                wps_start_count = 0;
 
                /* LEDs back to normal blink */
                if (wps_led_state)
                    send_wps_led_cmd(LED_BLINK_NORMAL, 0);
                else
                    send_wps_led_cmd(LED_BLINK_OFF, 0);

                /* Use previous ant pattern */
                wps_led_pattern = convert_ant_status_to_LED_pattern(ant_status);
                send_wps_led_cmd(LED_BLINK_PATTERN, wps_led_pattern);
            }
        }
        else
        {
            /* Check WPS button status */
            status = check_wps_button();

            if (status == ACTION_BLINK_TOGGLE)
            {
                wps_led_state == 1 ? 0 : 1 ;

                if (wps_led_state)
                    send_wps_led_cmd(LED_BLINK_NORMAL, 0);
                else
                    send_wps_led_cmd(LED_BLINK_OFF, 0);
            }
            else
            if (status  == ACTION_BLINK_WPS)
            {
                wps_start_count = loop_count;

                /* TODO: start WPS */
                
                /* All LEDs on/off at the same time */
                send_wps_led_cmd(LED_BLINK_PATTERN, 0xFF);

                send_wps_led_cmd(LED_BLINK_WPS, 0);
            }

            /* Update LED blink pattern every 5 secs */
            loop_count++;
            if (loop_count % 20 && wps_start_count == 0)
            {
                ant_status = check_ant_status();
                wps_led_pattern = convert_ant_status_to_LED_pattern(ant_status);
                printf("ant: %d, pattern: %x\n", ant_status, wps_led_pattern);
                send_wps_led_cmd(LED_BLINK_PATTERN, wps_led_pattern);
            }
        }

        usleep(250000);
    }
}
